//! Ported to Rust from <https://www.shadertoy.com/view/Ms2SD1>
//!
//! Original comment:
//! ```glsl
//! /*
//! * "Seascape" by Alexander Alekseev aka TDM - 2014
//! * License Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.
//! * Contact: tdmaav@gmail.com
//! */
//! ```
#![cfg_attr(target_arch = "spirv", no_std)]
#![feature(lang_items)]
#![feature(register_attr)]
#![register_attr(spirv)]

use core::f32::consts::PI;
use core::ops::{Add, Mul, Sub};

use spirv_std::glam::{const_mat2, const_vec3, Mat2, Mat3, Vec2, Vec3, Vec3Swizzles, Vec4Swizzles, Vec4};
use spirv_std::storage_class::{Input, Output, UniformConstant};

// Note: This cfg is incorrect on its surface, it really should be "are we compiling with std", but
// we tie #[no_std] above to the same condition, so it's fine.
#[cfg(target_arch = "spirv")]
use spirv_std::num_traits::Float;

pub fn saturate(x: f32) -> f32 {
    x.max(0.0).min(1.0)
}

pub fn pow(v: Vec3, power: f32) -> Vec3 {
    Vec3::new(v.x.powf(power), v.y.powf(power), v.z.powf(power))
}

pub fn exp(v: Vec3) -> Vec3 {
    Vec3::new(v.x.exp(), v.y.exp(), v.z.exp())
}

/// Based on: https://seblagarde.wordpress.com/2014/12/01/inverse-trigonometric-functions-gpu-optimization-for-amd-gcn-architecture/
pub fn acos_approx(v: f32) -> f32 {
    let x = v.abs();
    let mut res = -0.155972 * x + 1.56467; // p(x)
    res *= (1.0f32 - x).sqrt();

    if v >= 0.0 {
        res
    } else {
        PI - res
    }
}

pub fn smoothstep(edge0: f32, edge1: f32, x: f32) -> f32 {
    // Scale, bias and saturate x to 0..1 range
    let x = saturate((x - edge0) / (edge1 - edge0));
    // Evaluate polynomial
    x * x * (3.0 - 2.0 * x)
}

pub fn mix<X: Copy + Mul<A, Output = X> + Add<Output = X> + Sub<Output = X>, A: Copy>(
    x: X,
    y: X,
    a: A,
) -> X {
    x - x * a + y * a
}

pub trait Clamp {
    fn clamp(self, min: Self, max: Self) -> Self;
}

impl Clamp for f32 {
    fn clamp(self, min: Self, max: Self) -> Self {
        self.max(min).min(max)
    }
}

impl Clamp for Vec3 {
    fn clamp(self, min: Self, max: Self) -> Self {
        self.max(min).min(max)
    }
}

impl Clamp for Vec4 {
    fn clamp(self, min: Self, max: Self) -> Self {
        self.max(min).min(max)
    }
}

pub trait FloatExt {
    fn gl_fract(self) -> Self;
    fn rem_euclid(self, rhs: Self) -> Self;
    fn gl_sign(self) -> Self;
    fn deg_to_radians(self) -> Self;
}

impl FloatExt for f32 {
    fn gl_fract(self) -> f32 {
        self - self.floor()
    }

    fn rem_euclid(self, rhs: f32) -> f32 {
        let r = self % rhs;
        if r < 0.0 {
            r + rhs.abs()
        } else {
            r
        }
    }

    fn gl_sign(self) -> f32 {
        if self < 0.0 {
            -1.0
        } else if self == 0.0 {
            0.0
        } else {
            1.0
        }
    }

    fn deg_to_radians(self) -> f32 {
        PI * self / 180.0
    }
}

pub trait VecExt {
    fn gl_fract(self) -> Self;
    fn sin(self) -> Self;
    fn cos(self) -> Self;
    fn powf_vec(self, p: Self) -> Self;
    fn sqrt(self) -> Self;
    fn ln(self) -> Self;
    fn rem_euclid(self, m: f32) -> Self;
    fn step(self, other: Self) -> Self;
    fn reflect(self, normal: Self) -> Self;
    fn distance(self, other: Self) -> f32;
    fn gl_sign(self) -> Self;
}

fn step(edge: f32, x: f32) -> f32 {
    if x < edge {
        0.0
    } else {
        1.0
    }
}

impl VecExt for Vec2 {
    fn gl_fract(self) -> Vec2 {
        Vec2::new(self.x.gl_fract(), self.y.gl_fract())
    }

    fn sin(self) -> Vec2 {
        Vec2::new(self.x.sin(), self.y.sin())
    }

    fn cos(self) -> Vec2 {
        Vec2::new(self.x.cos(), self.y.cos())
    }

    fn powf_vec(self, p: Vec2) -> Vec2 {
        Vec2::new(self.x.powf(p.x), self.y.powf(p.y))
    }

    fn sqrt(self) -> Vec2 {
        Vec2::new(self.x.sqrt(), self.y.sqrt())
    }

    fn ln(self) -> Vec2 {
        Vec2::new(self.x.ln(), self.y.ln())
    }

    fn rem_euclid(self, m: f32) -> Vec2 {
        Vec2::new(self.x.rem_euclid(m), self.y.rem_euclid(m))
    }

    fn step(self, other: Vec2) -> Vec2 {
        Vec2::new(step(self.x, other.x), step(self.y, other.y))
    }

    fn reflect(self, normal: Vec2) -> Vec2 {
        self - 2.0 * normal.dot(self) * normal
    }

    fn distance(self, other: Vec2) -> f32 {
        (self - other).length()
    }

    fn gl_sign(self) -> Vec2 {
        Vec2::new(self.x.gl_sign(), self.y.gl_sign())
    }
}

impl VecExt for Vec3 {
    fn gl_fract(self) -> Vec3 {
        Vec3::new(self.x.gl_fract(), self.y.gl_fract(), self.z.gl_fract())
    }

    fn sin(self) -> Vec3 {
        Vec3::new(self.x.sin(), self.y.sin(), self.z.sin())
    }

    fn cos(self) -> Vec3 {
        Vec3::new(self.x.cos(), self.y.cos(), self.z.cos())
    }

    fn powf_vec(self, p: Vec3) -> Vec3 {
        Vec3::new(self.x.powf(p.x), self.y.powf(p.y), self.z.powf(p.z))
    }

    fn sqrt(self) -> Vec3 {
        Vec3::new(self.x.sqrt(), self.y.sqrt(), self.z.sqrt())
    }

    fn ln(self) -> Vec3 {
        Vec3::new(self.x.ln(), self.y.ln(), self.z.ln())
    }

    fn rem_euclid(self, m: f32) -> Vec3 {
        Vec3::new(
            self.x.rem_euclid(m),
            self.y.rem_euclid(m),
            self.z.rem_euclid(m),
        )
    }

    fn step(self, other: Vec3) -> Vec3 {
        Vec3::new(
            step(self.x, other.x),
            step(self.y, other.y),
            step(self.z, other.z),
        )
    }

    fn reflect(self, normal: Vec3) -> Vec3 {
        self - 2.0 * normal.dot(self) * normal
    }

    fn distance(self, other: Vec3) -> f32 {
        (self - other).length()
    }

    fn gl_sign(self) -> Vec3 {
        Vec3::new(self.x.gl_sign(), self.y.gl_sign(), self.z.gl_sign())
    }
}

pub struct Inputs {
    pub resolution: Vec3,
    pub time: f32,
    pub mouse: Vec4,
}

const NUM_STEPS: usize = 8;
const _EPSILON: f32 = 1e-3;
impl Inputs {
    fn epsilon_nrm(&self) -> f32 {
        0.1 / self.resolution.x
    }
}
const AA: bool = true;

// sea
const ITER_GEOMETRY: usize = 3;
const ITER_FRAGMENT: usize = 5;
const SEA_HEIGHT: f32 = 0.6;
const SEA_CHOPPY: f32 = 4.0;
const SEA_SPEED: f32 = 0.8;
const SEA_FREQ: f32 = 0.16;
const SEA_BASE: Vec3 = const_vec3!([0.0, 0.09, 0.18]);
// const SEA_WATER_COLOR: Vec3 = const_vec3!([0.8, 0.9, 0.6]) * 0.6;
const SEA_WATER_COLOR: Vec3 = const_vec3!([0.8 * 0.6, 0.9 * 0.6, 0.6 * 0.6]);


fn sea_time(time : f32) -> f32 {
    1.0 + time * SEA_SPEED
}
    
const OCTAVE_M: Mat2 = const_mat2!([1.6, 1.2, -1.2, 1.6]);

// math
fn from_euler(ang: Vec3) -> Mat3 {
    let a1: Vec2 = Vec2::new(ang.x.sin(), ang.x.cos());
    let a2: Vec2 = Vec2::new(ang.y.sin(), ang.y.cos());
    let a3: Vec2 = Vec2::new(ang.z.sin(), ang.z.cos());
    Mat3::from_cols(
        Vec3::new(
            a1.y * a3.y + a1.x * a2.x * a3.x,
            a1.y * a2.x * a3.x + a3.y * a1.x,
            -a2.y * a3.x,
        ),
        Vec3::new(-a2.y * a1.x, a1.y * a2.y, a2.x),
        Vec3::new(
            a3.y * a1.x * a2.x + a1.y * a3.x,
            a1.x * a3.x - a1.y * a3.y * a2.x,
            a2.y * a3.y,
        ),
    )
}
fn hash(p: Vec2) -> f32 {
    let h: f32 = p.dot(Vec2::new(127.1, 311.7));
    (h.sin() * 43758.5453123).gl_fract()
}
fn noise(p: Vec2) -> f32 {
    let i: Vec2 = p.floor();
    let f: Vec2 = p.gl_fract();
    let u: Vec2 = f * f * (Vec2::splat(3.0) - 2.0 * f);
    -1.0 + 2.0
        * mix(
            mix(
                hash(i + Vec2::new(0.0, 0.0)),
                hash(i + Vec2::new(1.0, 0.0)),
                u.x,
            ),
            mix(
                hash(i + Vec2::new(0.0, 1.0)),
                hash(i + Vec2::new(1.0, 1.0)),
                u.x,
            ),
            u.y,
        )
}

// lighting
fn diffuse(n: Vec3, l: Vec3, p: f32) -> f32 {
    (n.dot(l) * 0.4 + 0.6).powf(p)
}
fn specular(n: Vec3, l: Vec3, e: Vec3, s: f32) -> f32 {
    let nrm: f32 = (s + 8.0) / (PI * 8.0);
    (e.reflect(n).dot(l).max(0.0)).powf(s) * nrm
}

// sky
fn get_sky_color(mut e: Vec3) -> Vec3 {
    e.y = (e.y.max(0.0) * 0.8 + 0.2) * 0.8;
    Vec3::new((1.0 - e.y).powf(2.0), 1.0 - e.y, 0.6 + (1.0 - e.y) * 0.4) * 1.1
}

// sea
fn sea_octave(mut uv: Vec2, choppy: f32) -> f32 {
    uv += Vec2::splat(noise(uv));
    let mut wv: Vec2 = Vec2::one() - uv.sin().abs();
    let swv: Vec2 = uv.cos().abs();
    wv = mix(wv, swv, wv);
    (1.0 - (wv.x * wv.y).powf(0.65)).powf(choppy)
}

fn map(p: Vec3, time : f32) -> f32 {
    let mut freq: f32 = SEA_FREQ;
    let mut amp: f32 = SEA_HEIGHT;
    let mut choppy: f32 = SEA_CHOPPY;
    let mut uv: Vec2 = p.xz();
    uv.x *= 0.75;

    let mut d: f32;
    let mut h: f32 = 0.0;

    let mut i = 0;
    while i < ITER_GEOMETRY {
        d = sea_octave((uv + Vec2::splat(sea_time(time))) * freq, choppy);
        d += sea_octave((uv - Vec2::splat(sea_time(time))) * freq, choppy);
        h += d * amp;
        let octave_m = OCTAVE_M;
        uv = octave_m.transpose() * uv;
        freq *= 1.9;
        amp *= 0.22;
        choppy = mix(choppy, 1.0, 0.2);
        i += 1;
    }
    p.y - h
}

fn map_detailed(p: Vec3, time : f32) -> f32 {
    let mut freq: f32 = SEA_FREQ;
    let mut amp: f32 = SEA_HEIGHT;
    let mut choppy: f32 = SEA_CHOPPY;
    let mut uv: Vec2 = p.xz();
    uv.x *= 0.75;
    let mut d: f32;
    let mut h: f32 = 0.0;
    let mut i = 0;
    while i < ITER_FRAGMENT {
        d = sea_octave((uv + Vec2::splat(sea_time(time))) * freq, choppy);
        d += sea_octave((uv - Vec2::splat(sea_time(time))) * freq, choppy);
        h += d * amp;
        let octave_m = OCTAVE_M;
        uv = octave_m.transpose() * uv;
        freq *= 1.9;
        amp *= 0.22;
        choppy = mix(choppy, 1.0, 0.2);
        i += 1;
    }
    p.y - h
}

fn get_sea_color(p: Vec3, n: Vec3, l: Vec3, eye: Vec3, dist: Vec3) -> Vec3 {
    let mut fresnel: f32 = (1.0 - n.dot(-eye)).clamp(0.0, 1.0);
    fresnel = fresnel.powf(3.0) * 0.5;

    let reflected: Vec3 = get_sky_color(eye.reflect(n));
    let refracted: Vec3 = SEA_BASE + diffuse(n, l, 80.0) * SEA_WATER_COLOR * 0.12;

    let mut color: Vec3 = mix(refracted, reflected, fresnel);
    let atten: f32 = (1.0 - dist.dot(dist) * 0.001).max(0.0);
    color += SEA_WATER_COLOR * (p.y - SEA_HEIGHT) * 0.18 * atten;

    color += Vec3::splat(specular(n, l, eye, 60.0));
    color
}

// tracing
fn get_normal(p: Vec3, eps: f32, time : f32) -> Vec3 {
    let mut n: Vec3 = Vec3::zero();
    n.y = map_detailed(p, time);
    n.x = map_detailed(Vec3::new(p.x + eps, p.y, p.z), time) - n.y;
    n.z = map_detailed(Vec3::new(p.x, p.y, p.z + eps), time) - n.y;
    n.y = eps;
    n.normalize()
}

fn height_map_tracing(ori: Vec3, dir: Vec3, p: &mut Vec3, time : f32) -> f32 {
    let mut tm: f32 = 0.0;
    let mut tx: f32 = 1000.0;
    let mut hx: f32 = map(ori + dir * tx, time);
    if hx > 0.0 {
        return tx;
    }
    let mut hm: f32 = map(ori + dir * tm, time);
    let mut tmid: f32 = 0.0;
    let mut i = 0;
    while i < NUM_STEPS {
        tmid = mix(tm, tx, hm / (hm - hx));
        *p = ori + dir * tmid;
        let hmid: f32 = map(*p, time);
        if hmid < 0.0 {
            tx = tmid;
            hx = hmid;
        } else {
            tm = tmid;
            hm = hmid;
        }
        i += 1;
    }
    tmid
}
fn epsilon_nrm(resolution : Vec2) -> f32 {
    0.1 / resolution.x
}
fn get_pixel(coord: Vec2, time : f32, resolution : Vec2) -> Vec3 {
    let mut uv: Vec2 = coord / resolution;
    uv = uv * 2.0 - Vec2::one();
    uv.x *= resolution.x / resolution.y;
    // ray
    let ang: Vec3 = Vec3::new((time * 3.0).sin() * 0.1, time.sin() * 0.2 + 0.3, time);
    let ori: Vec3 = Vec3::new(0.0, 3.5, time * 5.0);
    let mut dir: Vec3 = uv.extend(-2.0).normalize();
    dir.z += uv.length() * 0.14;
    dir = from_euler(ang).transpose() * dir.normalize();
    // tracing
    let mut p: Vec3 = Vec3::zero();
    height_map_tracing(ori, dir, &mut p, time);
    let dist: Vec3 = p - ori;
    let n: Vec3 = get_normal(p, dist.dot(dist) * epsilon_nrm(resolution), time);
    let light: Vec3 = Vec3::new(0.0, 1.0, 0.8).normalize();
    // color
    mix(
        get_sky_color(dir),
        get_sea_color(p, n, light, dir, dist),
        smoothstep(0.0, -0.02, dir.y).powf(0.2),
    )
}

// main
pub fn main_image(frag_coord: Vec2, time : f32, resolution : Vec2, mouse : Vec2) -> Vec4 {
    let time: f32 = time * 0.3 + mouse.x * 0.01;
    let mut color: Vec3;
    if AA {
        color = Vec3::zero();
        let mut i = -1;
        while i <= 1 {
            let mut j = -1;
            while j <= 1 {
                let uv: Vec2 = frag_coord + Vec2::new(i as f32, j as f32) / 3.0;
                color += get_pixel(uv, time, resolution);
                j += 1;
            }
            i += 1;
        }
        color /= 9.0;
    } else {
        color = get_pixel(frag_coord, time, resolution);
    }
    // post
    color.powf(0.65).extend(1.0)
}


#[allow(unused_attributes)]
#[spirv(fragment)]
pub fn main_fs(
	#[spirv(frag_coord)] frag_coord : Input<Vec4>,
	u_time : UniformConstant<f32>,
	u_resolution : UniformConstant<Vec2>,
	u_mouse : UniformConstant<Vec4>,
	mut output : Output<Vec4>
) {
	let coord = frag_coord.load();
	let res = u_resolution.load();
	let mouse = u_mouse.load();
	
    output.store(main_image(coord.xy(), u_time.load(), res, mouse.xy()))
}